﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using Commons;

namespace Zerich.Micex
{
    /// <summary>
    /// Класс для разбора типизированных данных из байтового потока, присылаемого шлюзом ММВБ
    /// </summary>
    internal class ByteArrayReader
    {
        private byte[] _arr;
        private int _offset;
        private Encoding _encoding;

        internal ByteArrayReader(byte[] arr)
        {
            _arr = arr;
            _offset = 0;
            _encoding = Encoding.Default;
        }

        internal int ReadInt()
        {
            var union = new IntBytes(_arr, _offset);
            _offset += 4;
            return union.Integer;
        }

        internal byte ReadByte()
        {
            byte b = _arr[_offset];
            _offset++;
            return b;
        }

        internal string ReadString(int length)
        {
            return _encoding.GetString(ReadBytes(length));
        }

        /// <summary>
        /// Считаем, что в оставшейся части потока содержится строка, целиком её читаем
        /// </summary>
        internal string ReadStringToEnd()
        {
            return ReadString(_arr.Length - _offset);
        }

        /// <summary>
        /// Читаем строку, заканчивающуюся на \0
        /// </summary>
        internal String ReadStringZ(int knownLength)
        {
            int zero = Array.IndexOf(_arr, (byte)0, _offset);
            var stringLength = (zero != -1) ? zero - _offset : knownLength;
            String s = _encoding.GetString(_arr, _offset, stringLength);
            if (knownLength != 0)
            {
                _offset += knownLength;
            }
            else
            {
                _offset += (stringLength + 1);
            }
            return s;
        }

        [StructLayout(LayoutKind.Explicit)]
        struct IntBytes
        {
            [FieldOffset(0)]
            public int Integer;
            [FieldOffset(0)]
            public byte Byte0;
            [FieldOffset(1)]
            public byte Byte1;
            [FieldOffset(2)]
            public byte Byte2;
            [FieldOffset(3)]
            public byte Byte3;

            public IntBytes(byte[] arr, int offset)
            {
                Integer = 0;
                Byte0 = arr[offset + 0];
                Byte1 = arr[offset + 1];
                Byte2 = arr[offset + 2];
                Byte3 = arr[offset + 3];
            }
        }

        public TimeSpan ReadTime()
        {
            int time = ReadInt();
            int hours = time/10000;
            int mins = (time/100)%100;
            int secs = time%100;
            return new TimeSpan(hours, mins, secs);
        }

        public DateTime ReadDate()
        {
            int date = ReadInt();
            int day = date/1000000;
            int month = (date/10000)%100;
            int year = date%10000;
            return new DateTime(year, month, day);
        }

        public byte[] ReadBytes(int length)
        {
            var buf = new byte[length];
            Array.Copy(_arr, _offset, buf, 0, length);
            _offset += length;
            return buf;
        }

        /// <summary>
        /// Выводим остаток потока в виде байтового массива.
        /// Трюк используется в том случае, когда ожидается строка в кодировке, отличной от
        /// кодировки по-умолчанию.
        /// </summary>
        /// <returns></returns>
        internal byte[] ReadBytesToEnd()
        {
            return ReadBytes(_arr.Length - _offset);
        }

        public char ReadChar()
        {
            return _encoding.GetChars(new byte[] {ReadByte()})[0];                
        }
    }
}

// Copyright (c) 2009 Automated Intelligence Systems. 
//   All the source codes are property of Automated Intelligence Systems and strictly for internal usage.
//   If you find this code please report sec@aintsys.com

// Zidar: Класс наполовину переделан под проект Octopus, добавлены кое-какие удобные функции.