﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using Commons;

namespace Zerich.Micex
{
    /// <summary>
    /// Перечень наборов полей, которые должны присутствовать в xml-файле конфигурации таблиц ММВБ (MicexTablesStructure.xml)
    /// </summary>
    public enum FieldsetName
    {
        OrderBookRequest,
        OrderBookResponse,
        SecuritiesRequest,
        SecuritiesResponse,
        OrdersResponse,
        OrderTransaction,
        Wd_Order_By_NumberTransaction,
        TradesResponse
    }

    class FieldsConverter
    {
        Dictionary<FieldsetName, Dictionary<string, Field>> sets = new Dictionary<FieldsetName, Dictionary<string, Field>>();
        Dictionary<Instrument, int> decimals_cache = new Dictionary<Instrument, int>();
        DateTime last_unknown_decimals_read = DateTime.MinValue;

        /// <summary>
        /// Перезагружает файл с конфигурацией таблиц преобразования для API Micex, обычно это micex_tables_structure.xml
        /// </summary>
        /// <exception cref="XmlConfigurationException"/>
        /// <exception cref="XmlException"/>
        public void Reload(string structure_file)
        {
            sets.Clear();
            XmlDocument doc = new XmlDocument();
            doc.Load(structure_file);

            foreach (XmlNode node in doc.GetElementsByTagName("Set"))
            {
                if (!(node is XmlElement))
                    continue;

                int checksum = 0;
                int index = 0;
                Dictionary<string, Field> list = new Dictionary<string, Field>();
                foreach (XmlNode n in node.ChildNodes)
                {
                    if (!(n is XmlElement) || (n as XmlElement).Name != "Field")
                        continue;

                    XmlElement e = n as XmlElement;
                    Field p = new Field
                    {
                        Index = index++,
                        Name = Utils.FindAttribute<string>("Name", e),
                        Type = Utils.FindAttribute<FieldType>("Type", e),
                        Length = Utils.FindAttribute<int>("Length", e)
                    };
                    list.Add(Utils.FindAttribute<string>("Name", e), p);
                    checksum += p.Length;
                }

                FieldsetName set_name = Utils.FindAttribute<FieldsetName>("Name", node as XmlElement);
                if (sets.ContainsKey(set_name))
                    throw new XmlConfigurationException("Имя " + set_name + " в конфигурации таблиц повторяется более одного раза");

                //Log.Push(LogItemType.Debug, "{0} -> {1} полей, длина {2}", set_name, list.Count, checksum);
                sets.Add(set_name, list);
            }
        }

        public Dictionary<string, Field> GetFieldset(FieldsetName name)
        {
            if (sets.ContainsKey(name))
                return sets[name];
            else
                return null;
        }
        
        /// <summary>
        /// Формирует строку для отсылки в шлюз из набора параметров и заданного правила формирования параметров
        /// </summary>
        /// <exception cref="ArgumentException"/>
        /// <exception cref="XmlConfigurationException"/>
        public string Encode(Dictionary<string, object> data, FieldsetName set_name)
        {
            if (!sets.ContainsKey(set_name))
                throw new ArgumentException(String.Format("Блок настройки параметров {0} отсутствует", set_name));

            List<Field> structure = new List<Field>(sets[set_name].Values);
            if (structure.IsNullOrEmpty())
                return null;    /// В настройке задан пустой набор полей, в MTE_OpenTable передаем пустую строку

            if (data.Count != structure.Count)
                throw new ArgumentException("Количество элементов в fields не соответствует количеству элементов в structure");

            string secboard = null;
            string seccode = null;

            StringBuilder builder = new StringBuilder();
            for (int i = 0; i < structure.Count; i++)
            {
                FieldType type = structure[i].Type;
                string name = structure[i].Name;
                if (type == FieldType.Char)
                {
                    if (!data.ContainsKey(name))
                        throw new ArgumentException("Задание на кодирование не содержит информации по параметру " + name.ToString());
                    if (!(data[name] is string))
                        throw new ArgumentException(String.Format("Тип параметра {0} ожидался как {1}, а на деле оказался {2}", i, typeof(char), structure[i].Type));

                    string text = data[name] as string;
                    text = text.PadRight(structure[i].Length, ' ').Substring(0, structure[i].Length);
                    builder.Append(text);

                    if (name.ToString() == "SECCODE")
                        seccode = text.Trim();
                    else if (name.ToString() == "SECBOARD")
                        secboard = text.Trim();
                }
                else if (type == FieldType.Integer)
                {
                    string text = ((int)data[name]).ToString().PadLeft(structure[i].Length, '0');
                    builder.Append(text);
                }
                else if (type == FieldType.Long)
                {
                    string text = ((long)data[name]).ToString().PadLeft(structure[i].Length, '0');
                    builder.Append(text);
                }
                else if (type == FieldType.Date)
                {
                    if (data[name] == null)
                        builder.Append("".PadLeft(structure[i].Length, ' '));
                    else
                    {
                        DateTime date = (DateTime)data[name];
                        string text = date.ToString("yyyyMMdd");
                        builder.Append(text);
                    }
                }
                else if (type == FieldType.Time)
                {
                    if (data[name] == null)
                        builder.Append("".PadLeft(structure[i].Length, ' '));
                    else
                    {
                        DateTime date = (DateTime)data[name];
                        string text = date.ToString("HHmmss");
                        builder.Append(text);
                    }
                }
                else if (type == FieldType.Price)
                {
                    double price = (double)data[name];
                    if (decimals_cache.Keys.Any(instr => instr.Ticker == seccode && instr.ClassCode == secboard))
                    {
                        int decimals = decimals_cache[decimals_cache.Keys.First(instr => instr.Ticker == seccode && instr.ClassCode == secboard)];
                        string text = ((int)(price * Math.Pow(10, decimals))).ToString().PadLeft(structure[i].Length, '0');
                        builder.Append(text);
                    }
                    else
                    {
                        Log.Push(String.Format("Значение decimals для {0}/{1} еще не определено", secboard, seccode));
                        builder.Append("".PadLeft(structure[i].Length, '0'));
                    }
                }
                else
                    throw new ArgumentException("Тип " + type.ToString() + " не поддерживается");
            }

            return builder.ToString();
        }

        /// <summary>
        /// Разбирает ответ шлюза согласно заданному набору правил разбора параметров
        /// </summary>
        /// <exception cref="ArgumentException"/>
        /// <exception cref="XmlConfigurationException"/>
        public object[] Decode(ByteArrayReader reader, FieldsetName name)
        {
            if (!sets.ContainsKey(name))
                throw new ArgumentException(String.Format("Блок настройки параметров {0} отсутствует", name));

            List<Field> structure = new List<Field>(sets[name].Values);

            byte fields_count = reader.ReadByte();
            int total_fields_length = reader.ReadInt();
            byte[] fields_numbers;
            object[] data = new object[structure.Count];

            if (fields_count == 0)
            {
                fields_count = (byte)structure.Count;
                fields_numbers = new byte[fields_count];
                for (byte i = 0; i < fields_count; i++)
                    fields_numbers[i] = i;
            }
            else
                fields_numbers = reader.ReadBytes(fields_count);

            string secboard = null;
            string seccode = null;

            for (int i = 0; i < fields_count; i++)
            {
                FieldType type = structure[fields_numbers[i]].Type;
                int length = structure[fields_numbers[i]].Length;
                string field_name = structure[fields_numbers[i]].Name;

                object field;
                string text = reader.ReadString(length);
                if (type == FieldType.Char)
                {
                    field = text.Trim();
                    if (field_name.ToString() == "SECCODE")
                        seccode = ((string)field).Trim();
                    else if (field_name.ToString() == "SECBOARD")
                        secboard = ((string)field).Trim();
                }
                else if (type == FieldType.Integer)
                {
                    if (length > 10)
                        throw new MicexApiException("FieldType.Integer не может быть длиннее 10 символов");

                    int value;
                    int.TryParse(text, out value);
                    field = value;
                }
                else if (type == FieldType.Long)
                {
                    if (length <= 10)
                        throw new MicexApiException("FieldType.Long не может быть короче 11 символов");

                    long value;
                    long.TryParse(text, out value);
                    field = value;
                }
                else if (type == FieldType.Fixed)
                {
                    int value;
                    int.TryParse(text, out value);

                    field = (double)value / 100;
                }
                else if (type == FieldType.Price)
                {
                    int decimals = 0;
                    if (decimals_cache.Keys.Any(instr => instr.Ticker == seccode && instr.ClassCode == secboard))
                    {
                        decimals = decimals_cache[decimals_cache.Keys.First(instr => instr.Ticker == seccode && instr.ClassCode == secboard)];
                        int value;
                        int.TryParse(text, out value);
                        field = (double)value / Math.Pow(10, decimals);
                    }
                    else
                    {
                        if ((DateTime.Now - last_unknown_decimals_read).TotalSeconds >= 10)
                        {
                            Log.Push(LogItemType.FileOnly, "Произведена попытка чтения поля ЦЕНЫ при неизвестном значении decimals");
                            last_unknown_decimals_read = DateTime.Now;
                        }
                        field = (double)0;
                    }
                }
                else if (type == FieldType.Date)
                {
                    try
                    {
                        field = DateTime.ParseExact(text, "yyyyMMdd", null);
                    }
                    catch
                    {
                        field = null;
                    }
                }
                else if (type == FieldType.Time)
                {
                    try
                    {
                        field = DateTime.ParseExact(text, "HHmmss", null);
                    }
                    catch
                    {
                        field = null;
                    }
                }
                else
                {
                    Log.Push(LogItemType.Error, "Тип " + type.ToString() + " не поддерживается");
                    field = null;
                }
                data[fields_numbers[i]] = field;
            }

            /// Записываем в кеш информацию о decimals
            if (name == FieldsetName.SecuritiesResponse)
            {
                var query =
                    from Field f in structure
                    where f.Name == "SECCODE"
                    select f.Index;
                int seccode_index = query.First();
                
                query =
                    from Field f in structure
                    where f.Name == "SECBOARD"
                    select f.Index;
                int secboard_index = query.First();

                query =
                    from Field f in structure
                    where f.Name == "DECIMALS"
                    select f.Index;
                int decimals_index = query.First();

                if (data[decimals_index] != null)
                {
                    seccode = (string)data[seccode_index];
                    secboard = (string)data[secboard_index];
                    int decimals = (int)data[decimals_index];
                    Instrument instrument = new Instrument { ClassCode = secboard, Ticker = seccode, GateId = 0 };
                    if (!decimals_cache.ContainsKey(instrument))
                        decimals_cache.Add(instrument, decimals);
                    else
                        decimals_cache[instrument] = decimals;
                }
            }

            return data;
        }
    }    

    public enum FieldType
    {
        Char,
        Integer,
        Long,
        Fixed,
        Price,
        Date,
        Time
    }

    public struct Field
    {
        public int Index;
        public string Name;
        public FieldType Type;
        public int Length;
    }
}
