﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Commons;
using System.Timers;
using System.Xml;
using OctopusServer.Market;
using System.Threading;
using System.Diagnostics;

namespace Zerich.Micex
{
    /// <summary>
    /// Драйвер шлюза ММВБ
    /// </summary>
    class MicexGate : Gate
    {
        MicexApi api;
        System.Timers.Timer refresh_timer;
        MicexSubType sub_type;
        
        public MicexGate(GateDescription desc)
            : base(desc)
        {
            sub_type = Utils.FindValue<MicexSubType>("SubType", desc.Content);

            // Запускаем Api
            api = new MicexApi(Utils.FindElement("Connections", desc.Content), Utils.FindValue<string>("MicexTablesStructureFilename", desc.Content), sub_type);

            // Таймер обновления котировок
            int interval = Utils.FindValue<int>("QuotesRefreshInterval", desc.Content);
            refresh_timer = new System.Timers.Timer(interval);
            refresh_timer.Elapsed += OnRefreshTimedEvent;
        }

        #region Gate members
        internal override GateState Connect()
        {
            state = GateState.Connecting;
            api.Connect();
            /// Подписываемся на котировки
            var query = from Instrument i in market.ConfiguredInstruments
                        where i.GateId == Id
                        select i;
            api.SubscribeQuotes(query);
            refresh_timer.Start();
            return state = GateState.Connected;
        }

        internal override GateState Disconnect()
        {
            state = GateState.Disconnecting;
            api.Disconnect();            
            return state = GateState.Disconnected;
        }

        void OnRefreshTimedEvent(object source, ElapsedEventArgs args)
        {
            if (state != GateState.Connected)
                return;

            UpdateQuotes();
            UpdateOtherTables();
        }

        void UpdateQuotes()
        {
            Stopwatch gate_stopwatch = Stopwatch.StartNew();
            Stopwatch internal_stopwatch = new Stopwatch();

            try
            {
                api.RefreshQuotes();
            }
            catch (MicexApiException e)
            {
                Log.Push(e, "Ошибка MicexAPI при обновлении котировок");
                state = GateState.Crashed;
                return;
            }
            catch (Exception e)
            {
                Log.Push(e, "Неизвестная ошибка шлюза ММВБ");
                state = GateState.Crashed;
                return;
            }
            gate_stopwatch.Stop();
            internal_stopwatch.Start();

            /// Обновляем котировки во внутренней базе
            while (!api.QuotesQueue.IsNullOrEmpty())
                UpdateQuotes(api.QuotesQueue.Dequeue(), gate_stopwatch, internal_stopwatch);
        }

        void UpdateOtherTables()
        {
            try
            {
                api.RefreshOtherTables();
            }
            catch (MicexApiException e)
            {
                Log.Push(e, "Ошибка MicexAPI при обновлении других таблиц");
                state = GateState.Crashed;
                return;
            }
            catch (Exception e)
            {
                Log.Push(e, "Неизвестная ошибка шлюза ММВБ");
                state = GateState.Crashed;
                return;
            }

            /// Обновляем securities
            while (api.SecuritiesUpdates.Count > 0)
            {
                MicexApi.MicesSecuritiesUpdate update = api.SecuritiesUpdates.Dequeue();
                market.SetInstrumentInfo(new Instrument { GateId = Id, ClassCode = update.SecBoard, Ticker = update.SecCode }, update.Info);
            }

            /// Обновляем ордера
            while (api.OrderUpdates.Count > 0)
            {
                bool need_register_order = false;
                Order order;
                MicexApi.MicexOrderUpdate update = api.OrderUpdates.Dequeue();

                OrderState new_state;
                if (update.Status == "O")
                    new_state = OrderState.Active;
                else if (update.Status == "W")
                    new_state = OrderState.Removed;
                else if (update.Status == "M")
                    new_state = OrderState.Executed;
                else
                    new_state = OrderState.Error;

                int new_residue = update.Balance;

                order = market.GetOrder(update.OrderId, Id);
                if (order == null)
                {
                    if (update.Price > 0)
                    {
                        order = new LimitOrder();
                        (order as LimitOrder).Price = update.Price;
                    }
                    else
                        order = new MarketOrder();

                    order.Account = Account;
                    order.ClientCode = update.ClientCode;
                    order.Comment = String.Empty;
                    order.Id = update.OrderId;
                    order.Instrument = new Instrument
                    {
                        ClassCode = update.SecBoard,
                        Ticker = update.SecCode,
                        GateId = Id
                    };
                    order.Lots = update.Quantity;
                    order.Operation = update.BuySell;
                    order.StrategyId = 0;
                    order.Registered = update.OrderTime;
                    order.Updated = DateTime.Now;

                    need_register_order = true;
                }
                else
                    if (order.State == OrderState.Registering)
                        need_register_order = true;

                if (need_register_order)
                    market.RegisterOrder(order);

                market.UpdateOrder(order, new_state, new_residue);
            }

            /// Обновляем сделки
            Dictionary<int, Deal> deals = market.GetDeals();
            while (api.TradeUpdates.Count > 0)
            {
                MicexApi.MicexTradeUpdate update = api.TradeUpdates.Dequeue();

                Deal deal = new Deal();
                deal.Id = update.TradeId;
                Instrument i = new Instrument { GateId = this.Id, ClassCode = update.SecBoard, Ticker = update.SecCode };
                deal.Instrument = i;

                if (deals.ContainsKey(deal.GetHashCode()))
                    continue;

                deal.OrderId = update.OrderId;
                deal.Date = update.TradeTime;
                deal.Operation = update.BuySell;
                deal.Account = update.Account;
                deal.ClientCode = update.ClientCode;
                deal.Instrument = new Instrument
                {
                    Ticker = update.SecCode,
                    ClassCode = update.SecBoard,
                    GateId = Id
                };
                deal.Price = update.Price;
                deal.Lots = update.Quantity;
                deal.Comment = update.Comment;
                deal.Commission = update.Commission;

                market.RegisterDeal(deal);
            }
        }

        internal override void PlaceOrder(Order order)
        {
            Stopwatch gate_stopwatch = Stopwatch.StartNew();
//            Log.Push("Отправляем заявку в шлюз ММВБ");
            order.Account = Account;

            if (order is LimitOrder)
            {
                Dictionary<string, object> parameters = new Dictionary<string, object>();

                if (sub_type == MicexSubType.Spot)
                {
                    parameters.Add("ACCOUNT", Account);
                    parameters.Add("BUYSELL", order.Operation == Side.Buy ? "B" : "S");
                    parameters.Add("MKTLIMIT", "L");
                    parameters.Add("SPLITFLAG", "S");
                    parameters.Add("IMMCANCEL", " ");
                    parameters.Add("PRICEYIELDENTERTYPE", "P");
                    parameters.Add("MMORDER", " ");
                    parameters.Add("SECBOARD", order.Instrument.ClassCode);
                    parameters.Add("SECCODE", order.Instrument.Ticker);
                    parameters.Add("PRICE", (order as LimitOrder).Price);
                    parameters.Add("QUANTITY", order.Lots);
                    parameters.Add("BROKERREF", order.ClientCode + " @ " + DateTime.Now.ToString(Config.PreciseTimeFormat));
                    parameters.Add("EXTREF", String.Empty);
                    parameters.Add("ACTIVATIONTIME", null);
                    parameters.Add("CLIENTCODE", order.ClientCode);
                }
                else if (sub_type == MicexSubType.Futures)
                {
                    parameters.Add("TRDACCID", order.ClientCode);
                    parameters.Add("BUYSELL", order.Operation == Side.Buy ? "B" : "S");
                    parameters.Add("MKTLIMIT", "L");
                    parameters.Add("SPLITFLAG", "S");
                    parameters.Add("IMMCANCEL", " ");
                    parameters.Add("MMORDER", " ");
                    parameters.Add("SECBOARD", order.Instrument.ClassCode);
                    parameters.Add("SECCODE", order.Instrument.Ticker);
                    parameters.Add("PRICE", (order as LimitOrder).Price);
                    parameters.Add("QUANTITY", order.Lots);
                    parameters.Add("BROKERREF", order.ClientCode + " @ " + DateTime.Now.ToString(Config.PreciseTimeFormat));
                    parameters.Add("EXTREF", String.Empty);
                }
                else
                    throw new NotImplementedException();

                string message;
                TimeSpan external_delay;
                MicexReturnCode code = api.ExecTransaction("ORDER", FieldsetName.OrderTransaction, parameters, out message, out external_delay);
                order.ExternalDelay = external_delay;

                if (code == MicexReturnCode.Ok)
                {
                    order.State = OrderState.Registering;
                    long id = order.Id;
                    if (long.TryParse(message, out id))
                        order.Id = id;
                }
                else if (code == MicexReturnCode.Tsmr)
                {
                    order.State = OrderState.Error;
                    order.Comment = "TSMR: фатальный сбой при отправке заявки в шлюз, скорее всего нет связи с сервером";
                }
                else
                {
                    order.State = OrderState.Refused;
                    order.Comment = code.ToString() + ", " + message;
                }
            }
            else
            {
                order.State = OrderState.Refused;
                order.Comment = "Размещение ордеров типа, отличного от LimitOrder через шлюз ММВБ не реализовано";
                Log.Push(order.Comment);
            }

            order.GateDelay = gate_stopwatch.Elapsed - (order.ExternalDelay ?? new TimeSpan());
        }

        internal override void CancelOrder(Order order)
        {
            Dictionary<string, object> parameters = new Dictionary<string, object>();
            parameters.Add("ORDERNO", order.Id);

            string message;
            TimeSpan gate_delay;
            MicexReturnCode code = api.ExecTransaction("WD_ORDER_BY_NUMBER", FieldsetName.Wd_Order_By_NumberTransaction, parameters, out message, out gate_delay);

            if (code == MicexReturnCode.Ok || code == MicexReturnCode.TransRejected)
                order.State = OrderState.Removed;
            else
                order.State = OrderState.Refused;

            order.Comment = code.ToString() + ", " + message;
        }
        #endregion

        /// <summary>
        /// Выдает объем позиции в лотах, в которую разрешено войти указанному клиенту
        /// по указанному инструменту и направлению сделки
        /// </summary>
        internal override int? GetTradeAmount(Client client, Instrument instrument, Side operation)
        {
            ClientInfo ci = market.GetClientInfo(client);
            if (ci == null)
            {
                Log.Push(LogItemType.Error, "Запрошено вычисление доступного торгового объема для несуществующего клиента: " + client.ClientCode);
                return null;
            }


            InstrumentConfiguration details = market.GetInstrumentConfiguration(instrument);
            if (details == null)
                return null;

            if (details.Currency != ci.Currency)
                throw new NotImplementedException();
            
            int instrument_amount = ci.GetInstrumentAmount(instrument);

            decimal? money = GetTradeMoney(client, operation, details.IsMargin, details.IsCover);
            if (money == null)
                return null;

            Glass glass = market.GetGlass(instrument);
            if (operation == Side.Buy)
            {
                if (glass == null || glass.Ask == null || glass.Ask == 0)
                    return null;
                return (int?)Math.Floor(money.Value / (decimal)glass.Ask.Value);
            }
            else
            {
                if (glass == null || glass.Bid == null || glass.Bid == 0)
                    return null;
                return (int?)Math.Floor(money.Value / (decimal)glass.Bid.Value) + Math.Max(instrument_amount, 0);
            }
        }

        private decimal? GetTradeMoney(Client client, Side operation, bool is_margin, bool is_cover)
        {
            ClientInfo ci = market.GetClientInfo(client);
            decimal leverage = is_margin ? (decimal)ci.Leverage : 1;
            decimal? shorts = market.GetShorts(client);
            decimal? longs = market.GetLongs(client);
            decimal? buy_margin_block = market.GetBuyMarginBlock(client);
            decimal? buy_non_margin_block = market.GetBuyNonMarginBlock(client);
            decimal? buy_cover_block = market.GetBuyCoverBlock(client);
            decimal? sell_block = market.GetSellBlock(client);
            if (shorts == null || longs == null || buy_margin_block == null || buy_non_margin_block == null || buy_cover_block == null || sell_block == null)
                return null;

            decimal? buy_block = buy_margin_block + buy_cover_block;
            decimal? current_limit = (decimal)ci.Leverage * (ci.Money - buy_non_margin_block + longs + shorts);
            decimal? dost_current_limit = current_limit + shorts - sell_block
                + Utils.Negative<decimal>(ci.Money - buy_non_margin_block.Value) + shorts - buy_block;

            if (operation == Side.Buy)
            {
                if (is_margin && is_cover)
                    return dost_current_limit + Utils.Positive<decimal>(ci.Money - buy_non_margin_block.Value) + shorts - buy_block;
                else if (is_cover)
                {
                    decimal? open_pos = -shorts + sell_block + market.GetLongs(client, true) + buy_margin_block;
                    return (ci.Money - buy_non_margin_block + longs + shorts) + Utils.Negative<decimal>(current_limit.Value - open_pos.Value) - market.GetLongs(client, false) - buy_cover_block;
                }
                else
                    return dost_current_limit;
            }
            else
            {
                return dost_current_limit;
            }
        }
    }

    enum MicexSubType
    {
        Spot,
        Futures
    }
}
